# -*- coding: utf-8 -*-
# Copyright © 2018 VMware, Inc.  All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause
#
# Based on
# http://stackoverflow.com/questions/12023112/determine-variable-type-within-django-template
#

"""
Custom template filters to support display of YAMLCONF values via HTML views.
"""

from django import template

register = template.Library()


@register.filter
def get_type(value):
    """
    Return the name of the type for a value.  This is used when displaying
    values via HTML templates.
    """
    return type(value).__name__



# -*- coding: utf-8 -*-
# Copyright © 2018 VMware, Inc.  All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause
"""
List the settings defined via the YAMLCONF module.
"""
from django_yamlconf import list_attrs
from django_yamlconf.management.commands import YCBaseCommand


class Command(YCBaseCommand):
    """
    Implementation class for the "yclist" Django management command.
    """

    def handle(self, *args, **options):
        """
        Handle, i.e., execute, the command given the command line arguments
        "args" and "options".
        """
        super(Command, self).handle(*args, **options)
        list_attrs(stream=self.stdout)



# -*- coding: utf-8 -*-
# Copyright © 2018, VMware, Inc.  All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause
"""
Update the system control files based on the current YAMLCONF settings values.
"""
from __future__ import unicode_literals

from django.conf import settings
from django_yamlconf import sysfiles
from django_yamlconf.management.commands import YCBaseCommand


class Command(YCBaseCommand):
    """
    Create the system control files needed to support application, e.g.,
    the Nginx config file (/etc/nginx/nginx.conf).  The control files
    created are all the template files under the directory defined by the
    setttings YAMLCONF_SYSFILES_DIR directory mapping the files to the
    corresponding file under the root directory, e.g.,
    "buildaudit/templates/sys/etc/nginx/nginx.conf" is
    used to create the system control file "/etc/nginx/nginx.conf".

    The control file templates are process by the standard Django template
    engine with all the YAMLCONF attributes available for substitution or
    control logic, e.g., instead of hardcoding the server name in a config
    file, the Django template subsitution "{{ SERVER_NAME }}" should be
    used.

    The list of attributes available via YAMLCONF can be listed using the
    "yclist" management command.
    """

    help = __doc__

    def __init__(self, *args, **kwargs):
        super(Command, self).__init__(*args, **kwargs)
        self.root = None
        self.create = None

    def add_arguments(self, parser):
        """
        Add command line options.
        """
        super(Command, self).add_arguments(parser)
        parser.add_argument(
            '--root', "-R",
            dest='root',
            default="",
            help='Root directory to install expanded files instead of "/"'
        )
        parser.add_argument(
            '--create', "-C",
            dest='create',
            default=False,
            action="store_true",
            help="Create the control files if they do not exist",
        )
        parser.add_argument(
            '--noop', "-N",
            dest='noop',
            default=False,
            action="store_true",
            help="No-op mode, print template output instead (debugging)"
        )

    def handle(self, *args, **options):
        """
        Implmentation of the command, simply walk the templates "sys" directory
        and use "render_to_string" using the YAMLCONF attributes to create
        the contents for the system control file.
        """
        super(Command, self).handle(*args, **options)
        sysfiles(options['create'], options['noop'], settings, options['root'])



# -*- coding: utf-8 -*-
# Copyright © 2018 VMware, Inc.  All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause
"""
Django YAMLCONF management commands.
"""
from __future__ import unicode_literals

import importlib
import logging
import six

from django.conf import settings
from django.core.management.base import BaseCommand

import django_yamlconf

logger = logging.getLogger(__name__)


def add_cmd_attrs(defines, methods):
    """
    Add attributes defined on the command line, via name=value of an
    application method, to the set of attributes available in templates.
    """
    django_yamlconf.add_attributes(
        settings,
        get_methods(methods),
        "**GENERATED**"
    )
    django_yamlconf.add_attributes(
        settings,
        get_defines(defines),
        "**CMDLINE**"
    )


def get_defines(defines):
    """
    Return a dictionary of command line defined attribute values, e.g.,

        -D A=B

    returns

        {'A': 'B'}
    """
    result = {}
    for key_value in defines or []:
        try:
            key, value = key_value.split("=", 1)
            result[key] = value
            logger.debug('Adding "%s" => "%s"', key, value)
        except ValueError:
            logger.warning(
                'Invalid command line attribute definition "%s"',
                key_value
            )
    return result


def get_methods(methods):
    """
    Return a dictionary of attributes defined via application methods.
    """
    result = {}
    for method in methods or []:
        try:
            mod_name, func_name = method.rsplit('.', 1)
            func = getattr(
                importlib.import_module(mod_name),
                func_name
            )
            method_attrs = func()
            for key, value in six.iteritems(method_attrs):
                result[key] = value
        except ValueError:
            logger.warning('"%s" does not name a packaged method', method)
        except ImportError as ex:
            logger.warning('Method module "%s" failure: %s', mod_name, ex)
        except TypeError:
            logger.warning('"%s" did not return a dictionary', method)
        except AttributeError as ex:
            logger.warning(
                'Method module "%s" has no "%s": %s',
                mod_name, func_name, ex
            )
    return result


class YCBaseCommand(BaseCommand):
    """
    Common funcitonality for the YAMLCONF management commands: common
    options and handling.
    """

    def __init__(self, *args, **kwargs):
        super(YCBaseCommand, self).__init__(*args, **kwargs)
        self.verbosity = 1

    def add_arguments(self, parser):
        """
        Add command line options.
        """
        parser.add_argument(
            '--define', '-D',
            dest='defines',
            action='append',
            help="Additional command line defined attributes (name=value)"
        )
        parser.add_argument(
            '--attribute', '-A',
            dest='attributes',
            action='append',
            help="Method returning dict of addtional key/value pairs",
        )

    def handle(self, *args, **options):
        """
        Implmentation of the command, simply walk the templates "sys" directory
        and use "render_to_string" using the YAMLCONF attributes to create
        the contents for the system control file.
        """
        self.verbosity = options['verbosity']
        attributes = getattr(settings, 'YAMLCONF_ATTRIBUTE_FUNCTIONS', [])
        if options['attributes']:
            attributes.extend(options['attributes'])
        add_cmd_attrs(options['defines'], attributes)



# -*- coding: utf-8 -*-
# Copyright © 2018-2019, VMware, Inc.  All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause
"""
YAMLCONF
========

Handle YAML based Django settings: load Django settings from YAML files
based on a Django project name.  See the README.rst for additional
information.
"""
from __future__ import unicode_literals

import codecs
import copy
import getpass
import logging
import os
import platform
import sys
import textwrap
import traceback
import six

__all__ = [
    'add_attributes',
    'defined_attributes',
    'explain',
    'list_attrs',
    'load',
    'sysfiles',
]

logger = logging.getLogger(__name__)

_APPEND_MARKER = ":append"
_DOC_MARKER = ":doc"
_HIDE_MARKER = ":hide"
_PREPEND_MARKER = ":prepend"
_RAW_MARKER = ":raw"
_YAMLCONF_ATTRIBUTES = "_YAMLCONF_ATTRIBUTES"


def add_attributes(settings, attributes, source):
    """
    Add a set of name value pairs to the set of attributes, e.g.,
    attributes defined on the command line for management commands. Since
    this occurs after Django has loaded the settings, this function
    *does not*, in general, change behaviour of Django. It is used to
    add attribute definitions from management command lines. While this
    does not impact the behaviour of Django, it does make the attributes
    available for use in templates for the ``ycsysfiles`` command.

    :param settings: the Django settings module
    :param attributes: the dictionary of name/values pairs to add
    :param source: the name for the source (displayed by ``ycexplain``)
    :return: `None`

    """
    cur_attributes = get_cached_attributes(settings)
    for key, value in six.iteritems(attributes):
        set_attr_value(cur_attributes, settings, source, key, value)
    expand_attribute_refs(cur_attributes)
    inject_attr(cur_attributes, settings)


def add_attr_info(attributes, name, value, source="**INTERNAL**"):
    """
    Initialize a new attribute definition in the internal dict used to
    store the attribute and history of settings from various YAML files.
    """
    attributes[name] = {
        'value': value,
        'evalue': None,
        'source': source,
        'doc': [],
        'history': [],
        'hide': ('PASSWORD' in name),
    }


def extend_value(name, cur_value, value, append=True):
    """
    Append or prepend (depending on append argument)  the "value" to an
    existing value.
    """
    if not cur_value:
        return value
    elif isinstance(cur_value, (list, tuple)):
        result = copy.deepcopy(list(cur_value))
        if not isinstance(value, list):
            value = [value]
        if append:
            result.extend(value)
        else:
            result = value + result
        return result
    elif isinstance(cur_value, dict):
        result = copy.deepcopy(cur_value)
        if not isinstance(value, dict):
            logger.error(
                'Cannot append "%s" to "%s", "%s" is not a dict',
                value,
                name,
                type(value).__name__
            )
        else:
            for key in value.keys():
                result[key] = value[key]
        return result
    else:
        return [cur_value, value]


def bootstrap_attributes(base_dir):
    """
    Create the initial attribute set for YAMLCONF.  This set includes values
    that would not be available via other methods in a YAML file.  These
    attributes are available for attribute references in other values, e.g.,

        LOG_DIR: "{BASE_DIR}/log"
    """
    result = {}
    add_attr_info(result, 'TOP_DIR', os.path.dirname(base_dir))
    add_attr_info(result, 'BASE_DIR', base_dir)
    add_attr_info(result, 'VIRTUAL_ENV', os.environ.get("VIRTUAL_ENV", None))
    add_attr_info(result, 'USER', getpass.getuser())

    for name in ['machine', 'node', 'processor', 'release', 'system']:
        method = getattr(platform, name)
        add_attr_info(result, 'OS_{0}'.format(name.upper()), method())

    add_attr_info(result, 'PYTHON', {
        'MAJOR': sys.version_info.major,
        'MINOR': sys.version_info.minor,
        'MICRO': sys.version_info.micro,
        'RELEASELEVEL': sys.version_info.releaselevel,
        'SERIAL': sys.version_info.serial,
    })

    return result


def defined_attributes(settings=None, template_use=False):
    """
    Return a dictionary giving attribute names and associated values.
    This dictionary can be used as the variables when rendering templates.
    This is the set attributed used used as the variables when rendering
    templates for the ``ycsysfiles`` command.

    :param settings: the Django settings module (this is optional,
        defaults to the settings modules used when loading)
    :param template_use: If the the set of attributes return needs to be
        used to process a template, in the dictionary returned,
        attribute keys are added for dictionary parents e.g., "DATABASES", if
        "DATABASES.default..." is a YAMLCONF defined attribute.  The usage
        without this option support the `yclist` management command.
    :return: a dictionary giving attribute names and associated values.
    :rtype: dict
    """
    attributes = get_cached_attributes(settings)
    if not attributes:
        return {}
    result = {key: attributes[key]['evalue'] for key in attributes.keys()}
    # We are updating the result, avoid RunTimeError by creating a
    # temporary list for the set of keys in the iteration here
    for key in [k for k in result.keys()]:
        if '.' in key:
            base_key = key.split('.')[0]
            if base_key not in result:
                result[base_key] = getattr(settings, base_key)
    return result


def dirtree_find(filename, start_dir):
    """
    Iterator generating the instances of the given file name from the
    start_dir upto the root directory of the file system, e.g., the
    sequence of files generated might be

        /u/mrohan/clients/curwork/lib/python/xmpl.yaml
        /u/mrohan/clients/xmpl.yaml
        /u/mrohan/xmpl.yaml
    """
    curdir = start_dir or os.getcwd()
    done = False
    while not done:
        test_file = os.path.join(curdir, filename)
        if os.access(test_file, os.R_OK):
            yield test_file
        newdir = os.path.dirname(curdir)
        done = (newdir == curdir)
        curdir = newdir


def expand_attribute_refs(attributes):
    """
    Expand references to attributes in values, e.g., the set of definitions

        BASE_DIR: /var/osstp
        WEBAPPS_DIR: "{BASE_DIR}/webapps"

    would result in the definitions

        base_dir: /var/osstp
        webapps_dir: /var/osstp/webapps

    As can be seen from this example, attribute references use the Python
    dictionary formatting syntax to refer to other attributes.

    The routine "expand_attr_helper" is used to recurse over the attribute
    data.
    The set of expanded attribute values created is used to set the 'evalue's
    for each attribute.
    """
    values = expand_attr_helper(
        {key: attributes[key]['value'] for key in attributes.keys()}
    )
    for key, value in six.iteritems(values):
        attributes[key]['evalue'] = value


def expand_attr_helper(value, name="root", wrt=None):
    """
    Recurse over the attributes loaded expanding references.

    The name parameter is used to generate a path name to the attribute
    in case of errors, gives the user a little more information on which
    attribute has issues.

    The "wrt" (with respect to) is the set of attribute definitions.  This
    routine recursively expands included dictionaries and lists.  While
    the "value" to expand traverses down the full attribute set, the "wrt"
    value is always the top level attribute set.
    """
    wrt = wrt or value
    if isinstance(value, dict):
        for key in value.keys():
            if not value.get("{0}{1}".format(key, _RAW_MARKER), False):
                value[key] = expand_attr_helper(
                    value[key],
                    "{}[{}]".format(name, key),
                    wrt
                )
    elif isinstance(value, list):
        for i, _ in enumerate(value):
            value[i] = expand_attr_helper(
                value[i],
                "{}[{}]".format(name, i),
                wrt
            )
    elif isinstance(value, six.string_types):
        n_expand = 0
        done = False
        while not done:
            try:
                new_value = value.format(**wrt)
                done = (new_value == value)
                value = new_value
            except ValueError:
                logger.error('Invalid format value "%s" for "%s"', value, name)
                done = True
            except KeyError:
                logger.error('Undefined attribute "%s" in "%s"', value, name)
                done = True
            n_expand += 1
            if n_expand >= 10:
                logger.error('Recursive definition for "%s"', name)
                done = True
    return value


def explain(name, settings=None, stream=None):
    """
    Explain the source for an attribute definition including sources that
    were eclipsed by higher level YAML definition files.  If the attribute
    has associated documentation, it is also printed.

    This routine is only used by the YAMLCONF management command ``ycexplain``.

    :param name: the YAMLCONF controlled setting name
    :param settings: the Django settings module
    :param stream: the stream to write the explanation text (defaults to
        ``sys.stdout``)
    :return: `None`
    """
    stream = stream or sys.stdout
    attr_info = get_attr_info(name, settings=settings)
    if attr_info is None:
        stream.write("The setting \"{0}\" is not managed by YAMLCONF\n".format(
            name
        ))
        return
    stream.write("---------------------------\n")
    stream.write("{0} = \"{1}\" (via \"{2}\")\n".format(
        name,
        attr_info['value'],
        attr_info['source']
    ))
    if attr_info['value'] != attr_info['evalue']:
        stream.write("{0:{1}} = \"{2}\"\n".format(
            "",
            len(name),
            attr_info['evalue']
        ))
    stream.write("\n")
    documentation = attr_info['doc']
    if documentation:
        stream.write("Documentation:\n")
        for docstring in documentation:
            stream.write("{0}\n\n".format("\n".join(textwrap.wrap(
                docstring,
                width=65,
                initial_indent="    ",
                subsequent_indent="    "
            ))))
    if attr_info['history']:
        stream.write("Eclipsed values:\n")
        for value, source in attr_info['history']:
            stream.write("    \"{0}\" via \"{1}\"\n".format(value, source))


def get_attr_info(name, settings=None):
    """
    Get the information an attribute.
    """
    attributes = get_cached_attributes(settings)
    return attributes.get(name, None)


def get_settings_value(settings, name):
    """
    Get the value of an attribute defined in the settings module.  Dict
    access named attributes, are handled by digging into the dict in the
    settings file.  (Dict access names contain "." characteres, e.g.,
    "DATABASES.default.NAME"
    """
    components = name.split(".")
    result = getattr(settings, components[0], None)
    for component in components[1:]:
        if not isinstance(result, dict):
            logger.error('Not a dictionary: "%s" at "%s"', name, component)
            return None
        if component in result:
            result = result[component]
        else:
            return None
    return result


def get_attr_data(attributes, settings, name):
    """
    Return the dict describing an attribute value loaded from a YAML file.
    If this attribute has not already been seen, a new entry is created for
    it and returned.
    """
    if name not in attributes:
        add_attr_info(
            attributes,
            name,
            get_settings_value(settings, name),
            getattr(settings, "__name__", "settings")
        )
    return attributes[name]


def get_cached_attributes(settings=None):
    """
    When the YAML config files have been loaded, the data associated with
    the attributes defined is "cached" in the settings module via the attribute
    _YAMLCONF_ATTRIBUTES, this routine returns this cached data.

    This routine is used to support the management commands for YAMLCONF.
    """
    if settings is None:
        from django.conf import settings
    if not hasattr(settings, _YAMLCONF_ATTRIBUTES):
        logger.error("No YAMLCONF attributes defined")
        logger.error('"django_yamlconf.load" forgotten in the settings file?')
        setattr(settings, _YAMLCONF_ATTRIBUTES, {})
    return getattr(settings, _YAMLCONF_ATTRIBUTES)


def get_loader(syntax):
    """
    Return the "load" routine associated with the given file syntax, e.g., the
    "yaml.load" function.  The file syntax is assumed to name a module that
    has a "load" routine which parse an open file and returns dict/list of the
    file contents, e.g., "yaml.load", "json.load".
    """
    loader = None
    kwargs = {}
    try:
        module = __import__(syntax)
        loader = module.load
        if hasattr(module, "FullLoader"):
            kwargs = {'Loader': getattr(module, "FullLoader")}
    except ImportError as ex:
        logger.error('Unsupported YAMLCONF format "%s": %s', syntax, ex)
    except AttributeError as ex:
        logger.error('Loader "%s" has no "load" method: %s', syntax, ex)
    return loader, kwargs


def get_settings():
    """
    Return the settings module for the Django app that called the
    "yamlconf_load".  This should have been called from within the settings
    module so the stack trace back should give us the this module 3 up the
    call stack ("settings" + "get_settings" + "yamlconf_load").  The stack
    track is used to access the file name associate with the caller (0'th
    element) and from this the directory name is used as the Django application
    name.  Simply return "appname.settings" from the loaded modules in
    "sys.modules".
    """
    return sys.modules[
        "{0}.settings".format(os.path.basename(os.path.dirname(
            traceback.extract_stack(limit=3)[0][0]
        )))
    ]


def get_settings_dir(settings):
    """
    Return the directory containing the settings.py file.  This is simply
    the __file__ attribute of the settings module.  However, for testing
    purposes (and maybe it's useful in other contexts?) if the settings
    module is actually a class, return the __file__ value for the module
    defining the class.
    """
    path = os.getcwd()
    if hasattr(settings, "__file__"):
        path = settings.__file__
    elif hasattr(settings, "__module__"):
        path = sys.modules[settings.__module__].__file__
    return os.path.dirname(path)


def inject_attr(attributes, settings):
    """
    When the attributes have been loaded from various YAML files, this
    routine will "inject" the values into the "settings" module via
    the standard "setattr".   If the attribute names a dict element, i.e.,
    contains "." characters, the "injection" needs to traverse/create
    the supporting dictionaries in the settings file.

    To support the YAMLCONF management commands, the data accumlated on
    the attributes loaded is "cached" in the settings module via the
    attribute name _YAMLCONF_ATTRIBUTES.
    """
    setattr(settings, _YAMLCONF_ATTRIBUTES, attributes)
    for attr in attributes.keys():
        value = attributes[attr]['evalue']
        if '.' in attr:
            # Nested attribute (elements of a dict in the settings file)
            components = attr.split(".")
            target = getattr(settings, components[0], {})
            for key in components[1:-1]:
                if key not in target:
                    target[key] = {}
                target = target[key]
            if isinstance(target, dict):
                target[components[-1]] = value
        elif ':' not in attr:
            # Attributes with colons (:raw, :hide, etc) are not injected into
            # the settings module
            setattr(settings, attr, value)


def list_attrs(settings=None, stream=None):
    """
    Write a list of attributes managed by YAMLCONF to the given stream
    (defaults to ``sys.stdout``).
    Additional information can be printed using the ``explain`` routine.

    This routine is only used by the YAMLCONF management command ``yclist``.

    :param settings: the Django settings module
    :param stream: the stream to write the list text
    :return: `None`

    """
    stream = stream or sys.stdout
    attributes = get_cached_attributes(settings)
    if not attributes:
        stream.write("No YAMLCONF atttributes defined\n")
        return
    stream.write("Listing YAMLCONF managed attributes\n\n")
    # Determine the max attribute name length
    keylen = 0
    for name in attributes.keys():
        keylen = max(keylen, len(name))
    for name in sorted(attributes.keys()):
        stream.write("{0:<{1}}   {2}\n".format(
            name,
            keylen,
            attributes[name]['value']
        ))
    stream.write(
        "\nUse \"ycexplain\" for more information on individual attributes\n"
    )


def load(syntax="yaml", settings=None, base_dir=None, project=None):
    """
    Load the set of YAML files for a Django project.  The simplest usage is
    to call this at the end of a settings file.  In this context, no arguments
    are needed.

    :param syntax: The "syntax" parameter should name a Python module with
        a "load" method, e.g., the default is "yaml.load".  Other
        possibiliities could be "json" to use JSON formatted file or,
        even, "pickle" but that would be strange.  The "syntax" name is
        also used as the file extension for the YAMLCONF files.
    :param settings: The "settings" should be module containing the Django
        settings.  This is determined from the call stack if no module
        is given.
    :param base_dir: The "base_dir" defines the starting directory for
        YAMLCONF files and defaults to the directory containing the
        settings module.
    :param project: The "project" is the name of the Django project and
        defaults to the name of the directory containing the settings modules.
    :return: `None`
    """
    loader, loader_kwargs = get_loader(syntax)
    if loader is None:
        return
    settings = settings or get_settings()
    settings_dir = get_settings_dir(settings)
    base_dir = base_dir or os.path.dirname(settings_dir)
    project = project or os.path.basename(settings_dir)
    attr_filename = "{0}.{1}".format(project, syntax)
    attributes = bootstrap_attributes(base_dir)
    for filename in dirtree_find(attr_filename, settings_dir):
        load_conffile(attributes, settings, loader, loader_kwargs, filename)
    final_conf = os.environ.get("YAMLCONF_CONFFILE", None)
    if final_conf:
        load_conffile(attributes, settings, loader, loader_kwargs, final_conf)
    load_envdefs(attributes, settings)
    expand_attribute_refs(attributes)
    inject_attr(attributes, settings)


def load_conffile(attributes, settings, loader, loader_kwargs, filename):
    """
    Load an individual YAML file.  The data loaded is merged into the
    current set of attributes via the "set_attr_value" routine.
    """
    with codecs.open(filename, "r", encoding="utf-8") as defs:
        try:
            data = loader(defs, **loader_kwargs)
        except Exception as ex:
            logger.error('Failed to load "%s": %s', filename, ex)
            return
    if not isinstance(data, dict):
        logger.error(
            '"%s" did not define a dictionary: %s',
            filename,
            type(data)
        )
        return
    for name, value in six.iteritems(data):
        set_attr_value(attributes, settings, filename, name, value)


def load_envdefs(attributes, settings):
    """
    Load YAMLCONF attribute definitions from the environment.
    """
    for name, value in six.iteritems(os.environ):
        if name.startswith("YAMLCONF_"):
            set_attr_value(
                attributes,
                settings,
                "**ENVIRONMENT**",
                name.replace("YAMLCONF_", ""),
                value
            )


def set_attr_value(attributes, settings, filename, name, value):
    """
    Set an attribute value loaded from a YAML file.  If the attribute name
    ends with the _DOC_MARKED, the value is the documentation string for the
    attribute, set the 'doc' key instead.

    For regular attributes, the current value is stored in the history of
    loaded values (displayed via the "explain" routine and the "ycexplain"
    management command).
    """
    if name.endswith(_DOC_MARKER):
        name = name[:-len(_DOC_MARKER)]
        attr_data = get_attr_data(attributes, settings, name)
        attr_data['doc'].insert(0, value)
    elif name.endswith(_HIDE_MARKER):
        name = name[:-len(_HIDE_MARKER)]
        attr_data = get_attr_data(attributes, settings, name)
        attr_data['hide'] = True
    elif name.endswith(_APPEND_MARKER):
        name = name[:-len(_APPEND_MARKER)]
        attr_data = get_attr_data(attributes, settings, name)
        attr_data['history'].insert(
            0,
            (copy.deepcopy(attr_data['value']), attr_data['source'])
        )
        attr_data['source'] = filename
        attr_data['value'] = extend_value(name, attr_data['value'], value)
    elif name.endswith(_PREPEND_MARKER):
        name = name[:-len(_PREPEND_MARKER)]
        attr_data = get_attr_data(attributes, settings, name)
        attr_data['history'].insert(
            0,
            (copy.deepcopy(attr_data['value']), attr_data['source'])
        )
        attr_data['source'] = filename
        attr_data['value'] = extend_value(
            name,
            attr_data['value'],
            value,
            append=False
        )
    else:
        attr_data = get_attr_data(attributes, settings, name)
        attr_data['history'].insert(
            0,
            (copy.deepcopy(attr_data['value']), attr_data['source'])
        )
        attr_data['source'] = filename
        attr_data['value'] = value


def sf_init_file(create, noop, attrs, dst_filepath, src_filepath, render=None):
    """
    Initialize a system file based on a template under the "sys"
    template directory.
    """
    if not render:
        from django.template.loader import render_to_string
        render = render_to_string
    logger.debug(
        'YCSYSFILES: Rendering "%s" to "%s"',
        src_filepath,
        dst_filepath,
    )
    dst_filepath = os.path.normpath(dst_filepath.format(**attrs))
    contents = render(src_filepath, attrs)
    if noop:
        logger.info('Update: "%s"', dst_filepath)
        logger.info('-' * 50)
        logger.info("%s", contents)
        logger.info('-' * 50)
        return
    try:
        dirpath = os.path.dirname(dst_filepath)
        if create and not os.path.isdir(dirpath):
            os.makedirs(dirpath)
        with codecs.open(dst_filepath, "wb", "utf-8") as dest_file:
            logger.info('Updating the system control file "%s"', dst_filepath)
            dest_file.write(contents)
    except (IOError, OSError):
        # Raise by either the makedirs or the open
        logger.debug(
            'Skipping non-writeable or missing system file: "%s"',
            dst_filepath
        )


def sysfiles(create, noop, settings, rootdir="", render=None):
    """
    Traverse the sys templates directory expanding files to the destination
    directory.

    :param create: the template files should be created, normally will only
        update files that already exist on the system and are writable.
    :param noop: no-op mode, print what would be done.
    :param settings: the Django settings module
    :param rootdir: the directory to create the system files, defaults to
        ``/``, i.e., the root file system.
    :param render: the rendering engine, if not given, defaults to Django's
        ``render_to_string``
    :return: `None`
    """
    attributes = defined_attributes(settings, template_use=True)
    templates_dir = attributes.get("YAMLCONF_SYSFILES_DIR", None)
    if templates_dir:
        td_len = len(templates_dir)
        for root, _, files in os.walk(templates_dir):
            dst_dir = "{0}{1}".format(rootdir, root[td_len:])
            for name in files:
                dst_path = os.path.join(dst_dir, name)
                src_path = os.path.join(root, name)
                sf_init_file(
                    create,
                    noop,
                    attributes,
                    dst_path,
                    src_path,
                    render
                )
    else:
        logger.error("No YAMLCONF_SYSFILES_DIR settings defined")


def _load_version():
    """
    Return the version information defined in the VERSION file.
    """
    with open(os.path.join(os.path.dirname(__file__), "VERSION"), "r") as ver:
        v_info = ver.readline()
    major, minor, patch = filter(lambda x: x != '', v_info.strip().split(" "))
    return major, minor, patch


V_MAJOR, V_MINOR, V_PATCH = _load_version()
VERSION = "{0}.{1}.{2}".format(V_MAJOR, V_MINOR, V_PATCH)



# -*- coding: utf-8 -*-
# Copyright © 2018 VMware, Inc.  All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause
"""
Explain the source for settings values loaded via the YAMLCONF module.
"""
from __future__ import unicode_literals

from django_yamlconf import explain
from django_yamlconf.management.commands import YCBaseCommand


class Command(YCBaseCommand):
    """
    Implementation class for the "ycexplain" Django management command.
    """

    def add_arguments(self, parser):
        """
        Add the command line options for "ycexplain"
        """
        super(Command, self).add_arguments(parser)
        parser.add_argument(
            'attribute',
            nargs="+",
            help="Attribute to explain"
        )

    def handle(self, *args, **options):
        """
        Handle, i.e., execute, the command given the command line arguments
        "args" and "options".
        """
        super(Command, self).handle(*args, **options)
        for name in options['attribute']:
            explain(name, stream=self.stdout)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# Copyright © 2018 VMware, Inc.  All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

import codecs
from os import path
from setuptools import setup

def readme():
    with codecs.open('README.md', 'r', 'utf-8') as f:
        return f.read()

def version():
    vers_path = path.join(
        path.dirname(__file__),
        "django_yamlconf",
        "VERSION"
    )
    with open(vers_path, "r") as fh:
        v_info = fh.readline()
    return ".".join(filter(lambda x: x != '', v_info.split(" ")))


setup(
    name='django-yamlconf',
    version=version(),
    author='Michael Rohan',
    author_email='mrohan@vmware.com',
    description='Define Django settings in local YAML (or JSON) files',
    long_description=readme(),
    license='BSD-2',
    platforms=['Any'],
    keywords=['django'],
    url='https://github.com/vmware/django-yamlconf',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        "Framework :: Django",
        "Framework :: Django :: 1.7",
        "Framework :: Django :: 2.0",
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.5',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    install_requires=[
        'Django>=1.7',
        'PyYAML>=3.10',
        'six>=1.10.0',
    ],
    packages=[
        'django_yamlconf',
        'django_yamlconf.templatetags',
        'django_yamlconf.management',
        'django_yamlconf.management.commands',
    ],
    include_package_data=True,
    zip_safe=False,
    test_suite='nose.collector',
    tests_require=['nose', 'Jinja2'],
)



# -*- coding: utf-8 -*-
# Copyright © 2018 VMware, Inc.  All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause

"""
URL mappings to display the YAMLCONF settings.  This should only be used
for admin users.  The views are protected via

    @login_required(login_url='/admin/login/')

If you have a special environment, you might need to adjust this.
"""

from django.conf.urls import url
from django_yamlconf import views

app_name = 'django_yamlconf'
urlpatterns = [
    url("^$", views.index, name='index'),
    url("^(?P<name>.+)/$", views.attr_info, name='attr'),
]



# -*- encoding: utf-8 -*-
# Copyright © 2018 VMware, Inc.  All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause
"""
Basic views to allow browsing of the YAMLCONF definitions.
"""

import logging
from django.contrib.admin.views.decorators import staff_member_required
from django.http import Http404
from django.shortcuts import render_to_response
from django_yamlconf import get_cached_attributes
from django_yamlconf import get_attr_info


logger = logging.getLogger(__name__)


@staff_member_required
def index(request):
    """
    Generate the main page listing the YAMLCONF definitions.
    """
    title = 'YAMLCONF Attributes'
    logger.debug("Generating index page for YAMLCONF")
    return render_to_response(
        "yamlconf/index.html",
        {'attrs': get_cached_attributes(), 'title': title}
    )


@staff_member_required
def attr_info(request, name):
    """
    Display the page giving information on an individual attribute.
    """
    title = 'YAMLCONF: "{0}" Attribute'.format(name)
    logger.debug("Generating YAMLCONF info page for \"%s\"", name)
    info = get_attr_info(name)
    if info is None:
        logger.info("No such YAMLCONF attribute \"%s\"", name)
        raise Http404
    return render_to_response(
        "yamlconf/attribute.html",
        {'name': name, 'info': info, 'title': title}
    )
